//Simple Computer Design, expanded instruction set
module scomp2 (clock,reset,program_counter,register_A, stack_pointer, 
			  memory_address_register, memory_data_register, instruction_register);

input clock,reset;
output [7:0] program_counter,  memory_address_register, stack_pointer;
output [15:0] register_A, memory_data_register, instruction_register;

reg  [15:0] register_A,  instruction_register;
reg  [7:0] program_counter, stack_pointer;
reg  [7:0] state;

// State Encodings
parameter
		loadm    = 8'h00,
		load     = 8'h01,
		store    = 8'h02,
		reset_pc = 8'h80,
		fetch    = 8'h81,
		decode   = 8'h82,
		store2   = 8'h83,
		store3   = 8'h84;

reg [7:0] memory_address_register;
reg memory_write;
	
wire [15:0] memory_data_register;

// Use LPM function for computer's memory (256 16-bit words)
LPM_RAM_DQ	LPM_RAM_DQ_component (
				.address (memory_address_register),
				.inclock (clock),
				.data (register_A),
				.we (memory_write),
				.q (memory_data_register));
	defparam
		LPM_RAM_DQ_component.LPM_WIDTH = 16,
		LPM_RAM_DQ_component.LPM_WIDTHAD = 8,
		LPM_RAM_DQ_component.LPM_INDATA = "REGISTERED",
		LPM_RAM_DQ_component.LPM_ADDRESS_CONTROL = "UNREGISTERED",
		LPM_RAM_DQ_component.LPM_OUTDATA = "UNREGISTERED",
		LPM_RAM_DQ_component.USE_EAB = "ON",
// Reads in mif file for initial program and data values
		LPM_RAM_DQ_component.LPM_FILE = "program2.mif";


always @(posedge clock or posedge reset)
begin
	if (reset) state <= reset_pc;
    else 
	case (state)
	// reset the computer, need to clear some registers
    reset_pc :
    begin
		program_counter <= 8'h00;
		stack_pointer <= 8'hFF;
		memory_address_register <= 8'h00;
		register_A <= 16'h00;
		memory_write <= 0;
		state <= fetch;
	end
// Fetch instruction from memory and add 1 to program counter
	fetch :
   	begin
		instruction_register <= memory_data_register;
		program_counter <= program_counter + 1;
		memory_write <= 0;
		state <= decode;
   	end
// Decode instruction and send out address of any required data operands
   	decode :
   	begin
		memory_address_register <= instruction_register[7:0];
		state <= {1'b0,instruction_register[14:8]};
   	end
// Execute the LOADM instruction (use memory_data_register as address of data)
	loadm:
	begin
		memory_address_register <= memory_data_register[7:0];
		state <= load;
	end
// Execute the LOAD instruction
   	load :
   	begin
		register_A <= memory_data_register;
		memory_address_register <= program_counter;
		state <= fetch;
   	end
// Execute the STORE instruction (needs three clock cycles for memory write)
   	store :
   	begin
		// write register_A to memory
		memory_write <= 1;
		state <= store2;
   	end
// This state ensures that the memory address is valid until after memory_write goes low
    store2 :
   	begin
		memory_write <= 0;
		state <= store3;
   	end
   	store3 :
   	begin
		memory_address_register <= program_counter;
		state <= fetch;
   	end
// Execute the JUMP instruction
   	8'h03 :
   	begin
		memory_address_register <= instruction_register[7:0];
		program_counter <= instruction_register[7:0];
		state <= fetch;
   	end
// Execute the ADD instruction
	8'h04 :
   	begin
		register_A <= register_A + memory_data_register;
		memory_address_register <= program_counter;
		state <= fetch;
   	end
// Execute the SUB instruction
	8'h05 :
	begin
		register_A <= register_A - memory_data_register;
		memory_address_register <= program_counter;
		state <= fetch;
   	end
// Execute the JPOS instruction
	8'h09 :
	begin
		if (register_A[15]) memory_address_register <= program_counter;
		else 
		begin
			memory_address_register <= instruction_register[7:0];
			program_counter <= instruction_register[7:0];
		end
		state <= fetch;
	end
// Execute the JNEG instruction
	8'h0B :
	begin
		if (register_A[15])
		begin
			memory_address_register <= instruction_register[7:0];
			program_counter <= instruction_register[7:0];
		end
		else memory_address_register <= program_counter;
		state <= fetch;
	end
// Execute the LOADI instruction
	8'h11 :
	begin
		register_A <= {8'h00,instruction_register[7:0]};
		memory_address_register <= program_counter;
		state <= fetch;
	end
// Execute the ADDI instruction
	8'h12 :
	begin
		register_A <= register_A + instruction_register[7:0];
		memory_address_register <= program_counter;
		state <= fetch;
	end
// Undefined opcode
   	default :
   	begin
		memory_address_register <= program_counter;
		memory_write <= 0;
       	state <= fetch;
   	end
	endcase
end
endmodule

